from .horizons import Vector, Observer
from .animation import Body, play, play2d
from .plots import star_chart, plot_xyz, plot_altaz, plot_radec, Telescope
from .transform import radec_to_altaz, altaz_to_radec, radec_to_cartesian, to_xy_plane, rotating_coords
from .simbad import bright_objects, get_objects, explore_region, search_region, search_sky, explore_sky
from .catalogues import available_catalogues, Catalogue
from .cosmology import CosModel, Planck18

__version__ = "2.8.2"



"""
Module animation
================
This module supplies two functions, play() and play2d() to play animation, as well as
a class body to play user's defined bodies.
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from mpl_toolkits.mplot3d import Axes3D, proj3d

class Body:
    """
    User defined Body object to be entered in the play function.
    
    Parameters
    ----------
        name (str): name of the body
        pos (np.array): coordinates array with three columns (x, y, z)
                        shape of array is (n, 3); n: number of steps
        time (list): list of datetime objects; each element is the time
                     at which the coordinates is presented
    
    Attributes
    ----------
        x (np.array): x component of position vector
        y (np.array): y component of position vector
        z (np.array): z component of position vector
    """
    def __init__(self, name, pos, time):
        self.name = name
        self.name = name
        self.pos = pos
        self.time = time
        self.x = pos[:,0]
        self.y = pos[:,1]
        self.z = pos[:,2]

def play2d(bodies, names, colors, sizes, path=True, legend=True,
           interval=20, title=None, repeat=None):
    """
    Plays 2d animation from positions of several bodies.
    
    Arguments
    ---------
        bodies (list): targer bodies returned from Vector or Observer classes
        names (list): name of bodies
        colors (list): color of bodies
        sizes (list): size of bodies
        path (bool): whether or not draw the trajectory path of bodies;
                     default True.
        legend (bool): legend of the plot; default True.
        interval (int): time interval between sequences, greater means slower.
                        default 20.
        title (str) : title of the ax
        repeat (bool): whether or not repeat the animation
        

    Returns
    -------
        matplotlib.animation.FuncAnimation object
    """
    
    for b in bodies:
        if len(b.time)!=len(bodies[0].time):
            raise Exception('Number of steps are not equal for all bodies!')

    dates = bodies[0].time

    minxs = min([vec.x.min() for vec in bodies])
    minys = min([vec.y.min() for vec in bodies])
    maxxs = max([vec.x.max() for vec in bodies])
    maxys = max([vec.y.max() for vec in bodies])

    fig, ax = plt.subplots(figsize=(8,8))
    ax.set_aspect('equal')
    title = 'hypatie python package' if title is None else title
    ax.set_title(title)

    ax.set_xlabel('X')
    ax.set_ylabel('Y')


    minxs = minxs*1.2 if minxs<0 else minxs*0.8
    minys = minys*1.2 if minys<0 else minys*0.8
    maxxs = maxxs*0.8 if maxxs<0 else maxxs*1.2
    maxys = maxys*0.8 if maxys<0 else maxys*1.2
    
    ax.set_xlim([minxs, maxxs])
    ax.set_ylim([minys, maxys])
    


    txt = ax.text(maxxs-maxxs*1.2, maxys-maxys*.08, '')

    lines = []

    alpha = 0.2 if path else 0

    for i in range(len(bodies)):
        ax.plot(bodies[i].x, bodies[i].y, colors[i], alpha=alpha)
        lines.append(ax.plot(bodies[i].x, bodies[i].y,
                             color=colors[i], marker='o', markersize=sizes[i],
                             label=names[i])[0])

    def init():
        for line in lines:
            line.set_xdata(np.array([]))
            line.set_ydata(np.array([]))
        return lines

    def animate(i):
        for j,line in enumerate(lines):
            line.set_xdata(bodies[j].x[i])
            line.set_ydata(bodies[j].y[i])
        text_date = dates[i].isoformat()[:10]
        txt.set_text(text_date)
        return lines + [txt]

    if legend:
        plt.legend(loc='upper left')
    plt.grid(True)

    repeat = True if repeat is None else repeat

    anim = FuncAnimation(fig, animate, init_func=init,
                         frames=len(dates), interval=interval,
                         blit=True, repeat=repeat)
    return anim

def play(bodies, names, colors, sizes, path=True, legend=True, interval=20):
    """
    Plays animation from positions of several bodies.
    
    Arguments
    ---------
        bodies (list): targer bodies returned from Vector or Observer classes
        names (list): name of bodies
        colors (list): color of bodies
        sizes (list): size of bodies
        path (bool): whether or not draw the trajectory path of bodies;
                     default True.
        legend (bool): legend of the plot; default True.
        interval (int): time interval between sequences, greater means slower.
                        default 20.

    Returns
    -------
        matplotlib.animation.FuncAnimation object
    """
    
    for b in bodies:
        if len(b.time)!=len(bodies[0].time):
            raise Exception('Number of steps are not equal for all bodies!')

    dates = bodies[0].time

    minxs = min([vec.x.min() for vec in bodies])
    minys = min([vec.y.min() for vec in bodies])
    minzs = min([vec.z.min() for vec in bodies])
    maxxs = max([vec.x.max() for vec in bodies])
    maxys = max([vec.y.max() for vec in bodies])
    maxzs = max([vec.z.max() for vec in bodies])

    fig = plt.figure(figsize=plt.figaspect(0.5)*1.2)
    ax = fig.add_subplot(111, projection='3d')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.tick_params(axis='x', labelsize=8)
    ax.tick_params(axis='y', labelsize=8)
    ax.tick_params(axis='z', labelsize=8)

    txt = ax.text(minxs, maxys, maxzs, '')

    lines = []

    alpha = 0.2 if path else 0

    for i in range(len(bodies)):
        ax.plot(bodies[i].x, bodies[i].y, bodies[i].z, colors[i], alpha=alpha)
        lines.append(ax.plot(bodies[i].x[0:2], bodies[i].y[0:2], bodies[i].z[0:2],
                             color=colors[i], marker='o', markersize=sizes[i],
                             label=names[i])[0])

    def init():
        for line in lines:
            line.set_xdata(np.array([]))
            line.set_ydata(np.array([]))
            line.set_3d_properties(np.array([]))
        return lines

    def animate(i):
        for j,line in enumerate(lines):
            line.set_xdata(bodies[j].x[i])
            line.set_ydata(bodies[j].y[i])
            line.set_3d_properties(bodies[j].z[i])
        txt.set_text(dates[i].isoformat().replace('T', ' '))
        return lines + [txt]

    plt.locator_params(axis='x', nbins=7)
    plt.locator_params(axis='y', nbins=7)
    plt.locator_params(axis='z', nbins=7)

    if legend:
        plt.legend(loc='upper left')
    plt.grid(True)

    anim = FuncAnimation(fig, animate, init_func=init,
                         frames=len(dates), interval=interval,
                         blit=True, repeat=True)
    return anim



"""
Module horizons
===============
This module has two classes (Vector and Observer) to return positions
of the solar sytem objects by querying NASA's JPL Horizons API.

The Vector class provides positions with three dimensions: x,y,z.

The Observer class provides positions with two dimensions: ra,dec or
alt,az; depending on the request.
"""
import re
from datetime import datetime, timedelta
from urllib.request import urlopen
import numpy as np
from .plots import plot_altaz, plot_xyz

BASE_URL = 'https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&'

def _time_format(t): #check time format
    correct = True
    if isinstance(t, datetime):
        t = t.isoformat()[:19]
    elif isinstance(t, str) and bool(re.match("\d{4}-\d\d-\d\d \d\d:\d\d:\d\d", t)):
        t = t.replace(' ', 'T')
    else:
        correct = False
    return [correct, t]
    

class Vector:
    """
    Vector type coordinates (x, y, z)
    
    Parameters
    ----------
        target (str): targer body
        t1 (datetime or str): time from which to request data
        t2 (datetime or str): time to which to request data
        step (int): number of time intervals
        center (str): origin of coordinates. format: site@body
                      default '500@0'
        ref_plane (str): refrence plane: 'FRAME', 'ECLIPTIC' or 'BODY EQUATOR'
                         default 'FRAME' (equator plane)
        vec_table (int): vector table type: 1 for position; 2 for position and velocity
    
    Attributes
    ----------
        time (list): list of datetime objects; each element is the time
                     at which the coordinates is presented
        pos (np.array): coordinates array with three columns (x, y, z)
                        shape of array is (n, 3); n: number of steps
        x (np.array): x component of position vector
        y (np.array): y component of position vector
        z (np.array): z component of position vector
        vel (np.array): velocity array with three columns (vx, vy, vz)
                        shape of array is (n, 3); n: number of steps
        vx (np.array): x component of velocity vector
        vy (np.array): y component of velocity vector
        vz (np.array): z component of velocity vector
        url (str): url of NASA's JPL Horizons that produced the results
    
    Methods
    -------
        plot: plot position(s) (x,y,z) of the target body
    """

    def __init__(self, target, t1, t2=None, step=None, center='500@0', ref_plane='FRAME', vec_table=1):
        self.target = target
        self.center = center
        self.ref_plane = ref_plane
        self.vec_table = vec_table
        
        correct, t1 = _time_format(t1)
        if correct:
            self.t1 = t1
        else:
            raise Exception("t1 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
        self.init = False
        if t2 is None:
            self.init = True # user just wants the initial state
            step = 1
            t2 = datetime.strptime(self.t1, '%Y-%m-%dT%H:%M:%S') + timedelta(seconds=1)
        
        correct, t2 = _time_format(t2)
        if correct:
            self.t2 = t2
        else:
            raise Exception("t2 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
            
        self.step = step

        error_msg, time, pos, vel = self.get_request()
        
        if len(error_msg)==0:
            if self.init:
                self.time = time[0]
                self.pos = pos[0]
                self.x, self.y, self.z = pos[0][0], pos[0][1], pos[0][2]
                if vel.shape!=(0,):
                    self.vel = vel[0]
                    self.vx, self.vy, self.vz = vel[0,0], vel[0,1], vel[0,2]
                else:
                    self.vel = vel
                
            else:
                self.time = time
                self.pos = pos
                self.x, self.y, self.z = pos[:,0], pos[:,1], pos[:,2]
                self.vel = vel
                if vel.shape!=(0,):
                    self.vx, self.vy, self.vz = vel[:,0], vel[:,1], vel[:,2]
                else:
                    self.vx, self.vy, self.vz = vel, vel, vel
        else:
            raise Exception('\n'+error_msg[:-2])

    def vector_url(self, target, t1, t2, step, center, ref_plane, vec_table):
        params = f"""COMMAND='{target}'
        CENTER='{center}'
        MAKE_EPHEM='YES'
        TABLE_TYPE='VECTORS'
        START_TIME='{t1}'
        STOP_TIME='{t2}'
        STEP_SIZE='{step}'
        CAL_FORMAT='CAL'
        TIME_DIGITS='FRACSEC'
        OUT_UNITS='KM-S'
        REF_PLANE='{ref_plane}'
        REF_SYSTEM='J2000'
        VECT_CORR='NONE'
        VEC_LABELS='NO'
        VEC_DELTA_T='NO'
        CSV_FORMAT='YES'
        OBJ_DATA='NO'
        VEC_TABLE='{vec_table}'"""
        params = params.replace('\n', '&').replace(' ', '')
        url = BASE_URL + params
        self.url = url
        return url

    def get_request(self):
        error_msg = ''
        url = self.vector_url(self.target, self.t1, self.t2, self.step,
                              self.center, self.ref_plane, self.vec_table)
        with urlopen(url) as r:
            text = r.read().decode('utf-8')
        if ('$$SOE' not in text) or ('$$EOE' not in text):
            error_msg = text[:text.find('$$SOF')]
            return [error_msg, np.array([]), np.array([])]
        mark1 = text.find('$$SOE')
        text = text[mark1+6:]
        mark2 = text.find('$$EOE')
        text = text[:mark2]
        rows = text.split('\n')[:-1]
        n_columns = len(rows[0].split(',')[:-1])
        
        times = [i.split(',')[1].strip()[5:] for i in rows]
        times = [datetime.strptime(i, '%Y-%b-%d %H:%M:%S.%f') for i in times]

        # positions
        pS = [i.split(',')[2:5] for i in rows]
        pos_list = []
        for p in pS:
            tmp_ls = [float(i.strip())*1000 for i in p]
            pos_list.append(tmp_ls)

        # velocities
        if n_columns==8:
            vS = [i.split(',')[5:8] for i in rows]
            vel_list = []
            for v in vS:
                tmp_ls = [float(i.strip())*1000 for i in v]
                vel_list.append(tmp_ls)
        else:
            vel_list = []
        return [error_msg, times, np.array(pos_list), np.array(vel_list)]

    def plot(self, color='b', size=10):
        """
        plot position(s) (ra/dec or az/alt) of the target body
        
        Arguments
        ----------
            color (str): color of body
            size (int): size of body
        
        Returns
        -------
            matplotlib.axes.Axes object
        """
        if isinstance(self.time, list):
            return plot_xyz(self.x, self.y, self.z, color, size)
        else:
            return plot_xyz([self.x], [self.y], [self.z], color, size)


class Observer:
    """
    Observer type coordinates (ra/dec or az/alt)
    
    Parameters
    ----------
        ttarget (str): targer body
        t1 (datetime or str): time from which to request data
        t2 (datetime or str): time to which to request data
        step (int): number of time intervals
        center (str): origin of coordinates. format: site@body
                      default '500@0'
        quantities (int): 1 for ra/dec, 2 for az/alt; (default 2)
    
    Attributes
    ----------
        time (list): list of datetime objects; each element is the time
                     at which the coordinates is presented
        pos (np.array): coordinates array with two columns (ra/dec or az/alt)
                        shape of array is (n, 2); n: number of steps
        ra (np.array): right ascension OR azimuth (depending on quantities chosen)
        dec (np.array): declination OR altitude (depending on quantities chosen)
        url (str): url of NASA's JPL Horizons that produced the results
    
    Methods
    -------
        plot: plot position(s) (ra/dec or az/alt) of the target body
    """
    
    def __init__(self, target, t1, t2=None, step=None, center='500@399', quantities=2):
        self.target = target
        self.center = center
        self.quantities = quantities
        
        correct, t1 = _time_format(t1)
        if correct:
            self.t1 = t1
        else:
            raise Exception("t1 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
        self.init = False
        if t2 is None:
            self.init = True # user just wants the initial state
            step = 1
            t2 = datetime.strptime(self.t1, '%Y-%m-%dT%H:%M:%S') + timedelta(seconds=1)
        
        correct, t2 = _time_format(t2)
        if correct:
            self.t2 = t2
        else:
            raise Exception("t2 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
            
        self.step = step

        error_msg, time, pos = self.get_request()
        
        if len(error_msg)==0:
            if self.init:
                self.time = time[0]
                self.pos = pos[0]
                self.ra, self.dec = pos[0][0], pos[0][1]
            else:
                self.time = time
                self.pos = pos
                self.ra, self.dec = pos[:,0], pos[:,1]
        else:
            raise Exception('\n'+error_msg[:-2])

    def coord(self, center):        
        valid_center = True
        str_center = ''

        if len(center.split('@'))==2:
            body = center.split('@')[-1]
        elif len(center.split('@'))==1:
            body = '399'
        else:
            valid_center = False

        crd = center.split('@')[0]
        if len(crd.split(','))==3:
            crd = ','.join(crd.split(',')[:-1]) + ',' + str(int(crd.split(',')[-1])/1000)
        elif len(crd.split(','))==2:
            crd = crd + ',0'
        else:
            valid_center = False

        if valid_center:
            str_center = f"coord@{body}'&COORD_TYPE='GEODETIC'&SITE_COORD='{crd}"

        return [valid_center, str_center]
    
    def observer_url(self, target, t1, t2, step, center, quantities, skip_daylight=True):
        if len(center.split(',')) > 1: #coordinates
            valid_center, center = self.coord(center)
            skip_daylight = False
            quantities = 4 # apparent Ra & DEC
            if not valid_center:
                raise Exception('Center is not valid!')

        skip_daylight='YES' if skip_daylight==True else 'NO'
        
        params = f"""COMMAND='{target}'
        CENTER='{center}'
        MAKE_EPHEM='YES'
        TABLE_TYPE='OBSERVER'
        START_TIME='{t1}'
        STOP_TIME='{t2}'
        STEP_SIZE='{step}'
        CAL_FORMAT='CAL'
        TIME_DIGITS='FRACSEC'
        ANG_FORMAT='DEG'
        OUT_UNITS='KM-S'
        RANGE_UNITS='KM'
        APPARENT='AIRLESS'
        SUPPRESS_RANGE_RATE='YES'
        SKIP_DAYLT='{skip_daylight}'
        EXTRA_PREC='YES'
        R_T_S_ONLY='NO'
        REF_SYSTEM='J2000'
        CSV_FORMAT='YES'
        OBJ_DATA='NO'
        QUANTITIES='{quantities}'"""
        
        params = params.replace('\n', '&').replace(' ', '')
        url = BASE_URL + params
        self.url = url
        
        return url

    def get_request(self):
        error_msg = ''
        url = self.observer_url(self.target, self.t1, self.t2, self.step,
                                self.center, self.quantities)
        with urlopen(url) as r:
            text = r.read().decode('utf-8')
        if ('$$SOE' not in text) or ('$$EOE' not in text):
            error_msg = text[:text.find('$$SOF')]
            return [error_msg, np.array([]), np.array([])]
        mark1 = text.find('$$SOE')
        text = text[mark1+6:]
        mark2 = text.find('$$EOE')
        text = text[:mark2]
        rows = text.split('\n')[:-1]
        times = [i.split(',')[0].strip() for i in rows]
        times = [datetime.strptime(i, '%Y-%b-%d %H:%M:%S.%f') for i in times]
        pS = [i.split(',')[-3:-1] for i in rows]
        pos_list = []
        for p in pS:
            tmp_ls = [float(i.strip()) for i in p]
            pos_list.append(tmp_ls)
        return [error_msg, times, np.array(pos_list)]

    def plot(self, color='b', size=10):
        """
        plot position(s) (ra/dec or az/alt) of the target body
        
        Arguments
        ----------
            color (str): color of body
            size (int): size of body
        
        Returns
        -------
            matplotlib.axes.Axes object
        """
        plot_altaz(self.ra, self.dec, color, size)



"""
Module plot
================
This module supplies several functions for plotting reasons.
"""
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from collections.abc import Iterable
from urllib.request import urlopen
from .simbad import bright_objects
from .transform import radec_to_altaz, altaz_to_radec

def _equalize_scale(X,Y,Z, ax):
    """
    Equalize scales of x, y and z axis

    Arguments
    ---------
        X (np.array): x values
        Y (np.array): y values
        Z (np.array): z values
        ax (axes): axes object

    Returns
    -------
        matplotlib axes object
    """
    max_range = np.array([X.max()-X.min(), Y.max()-Y.min(), Z.max()-Z.min()]).max()
    Xb = 0.5*max_range*np.mgrid[-1:2:2,-1:2:2,-1:2:2][0].flatten() + 0.5*(X.max()+X.min())
    Yb = 0.5*max_range*np.mgrid[-1:2:2,-1:2:2,-1:2:2][1].flatten() + 0.5*(Y.max()+Y.min())
    Zb = 0.5*max_range*np.mgrid[-1:2:2,-1:2:2,-1:2:2][2].flatten() + 0.5*(Z.max()+Z.min())
    for xb, yb, zb in zip(Xb, Yb, Zb):
       ax.plot([xb], [yb], [zb], 'w')
    return ax

def plot_xyz(x, y, z, color, size, au=False):
    """
    Draw 3d cartesian plot of a body

    Arguments
    ---------
        x (np.array): x values
        y (np.array): y values
        z (np.array): z values
        color (str): color
        size (int): size
        au (bool): whether or not in AU unit; default False.

    Returns
    -------
        matplotlib axes object
    """
    fig = plt.figure(figsize=plt.figaspect(0.5)*1.2)
    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.tick_params(axis='x', labelsize=8)
    ax.tick_params(axis='y', labelsize=8)
    ax.tick_params(axis='z', labelsize=8)
    au = 149597870700.0 if au else 1
    ax.scatter(x/au, y/au, z/au, c=color, s=size)
    ax = _equalize_scale(x/au, y/au, z/au, ax)
    return ax

def plot_altaz(az, alt, mag=None, size=None, color='k', alpha=1, marker='o', ax=None):
    """
    Plot positions of bodies based on altitude/azimuth coordinates

    Arguments
    ---------
        az (iter of float): azimuth values
        alt (iter of float): altitude values
        mag (iter of float): apparent magnitudes; default None.
        size (int): size; default None.
        color (str): color; default 'k'.
        alpha (float): alpha value (transparency), between 0 and 1; default 1.
        marker (str): marker shape; default 'o'.
        ax (axes): axes object; default None.

    Returns
    -------
        matplotlib axes object
    """

    if isinstance(az, Iterable):
        az = np.array(az)
        alt = np.array(alt)
        mag = np.array(mag) if mag is not None else None
    else:
        az = np.array([az])
        alt = np.array([alt])
        mag = None
        
    az  = az*(np.pi/180)

    if size is None:
        if mag is None:
            size = [5] * len(az)
        else:
            size = (0.1 + max(mag)-mag)**1.8
    
    if ax is None:
        fig = plt.figure()
        ax = fig.add_axes([0.1,0.1,0.8,0.8], polar=True)
        ax.set_theta_zero_location('N')
        ax.set_rlim(90, 0, 1)
        ax.set_yticks(np.arange(0,91,30))
        ax.set_xticks([0, np.pi/4, np.pi/2, 3*np.pi/4, np.pi, 5*np.pi/4, 6*np.pi/4, 7*np.pi/4])
        ax.set_xticklabels(['N','NE','E','SE','S','SW','W','NW'])
        ax.tick_params(axis=u'both', which=u'both',length=0)
    if matplotlib.__version__ < '3.0.0':
        alt = [90-i for i in alt]
    ax.scatter(az, alt, c=color, s=size, alpha=alpha, marker=marker)
    ax.grid(True, alpha=0.7)
    return ax

def plot_radec(ra, dec, mag=None, size=None):
    """
    Plot positions of bodies based on RA/Dec coordinates

    Arguments
    ---------
        ra (iter of float): Right Ascension values
        dec (iter of float): Declination values
        mag (iter of float): apparent magnitudes; default None.
        size (int): size; default None.

    Returns
    -------
        matplotlib axes object
    """
        
    if isinstance(ra, Iterable):
        ra = np.array(ra)
        dec = np.array(dec)
        mag = np.array(mag) if mag is not None else None
    else:
        ra = np.array([ra])
        dec = np.array([dec])
        mag = None

    ra = (np.mod(ra-180, 360)-180) * (np.pi/180)
    dec = dec * (np.pi/180)

    if size is None:
        if mag is None:
            size = [5] * len(ra)
        else:
            size = (0.1 + max(mag)-mag)**1.8

    fig = plt.figure(figsize=(10,5))
    ax = fig.add_subplot(111, projection="aitoff")
    ax.grid(True)
    ax.scatter(ra, dec, s=size, c='k')
    plt.subplots_adjust(top=0.95,bottom=0.0)
    plt.xlabel('RA')
    plt.ylabel('Dec')
    return ax

def star_chart(lon, lat, t=None, otype=None):
    """
    Plot the star chart for a location and time on earth

    Arguments
    ---------
        lon (float): longitude of observer
        lat (float): latitude of observer
        t (datetime or str): time of observation; default None.
        otype (str): object type to be presented in the chart; default None.

    Returns
    -------
        matplotlib axes object
    """
    _, rows = bright_objects(otype)
    mag = [i[2] for i in rows]
    ra  = [i[3] for i in rows]
    dec = [i[4] for i in rows]
    alt, az = radec_to_altaz(lon, lat, ra, dec, t)
    ax = plot_altaz(az, alt, mag=mag)
    return ax

class Telescope:
    """
    Plot image of a location in the sky by querying NASA's Virtual
    Observatory (SkyView)

    Arguments
    ---------
        target_loc (tuple): location of a target point in the sky;
                            (ra, dec) or (az, alt)
        obs_loc (tuple): location of observer on Earth; format: (lon, lat)
        fov (float) : Field of View in degrees; default 1
        t (datetime or str): time of observation; default now
        survey (str): name of survey; default 'digitized sky survey'

    Attributes
    ----------
        data (numpy array): image data
        url (str) : image url

    Methods
    -------
        plot : plots the image and returns fig and ax
        show : shows the image
    """
    def __init__(self,
                 target_loc,
                 obs_loc=None,
                 fov=1,
                 t=None,
                 survey='digitized sky survey'):
        
        survey = survey.replace(' ', '+')

        if obs_loc is not None:
            lon, lat = obs_loc
            az , alt = target_loc
            ra, dec = altaz_to_radec(lon, lat, az, alt, t)
            ra, dec = float(ra), float(dec)
            position = str((ra, dec)).replace(' ', '')[1:-1]
        else:
            position = str(target_loc).replace(' ', '')[1:-1]

        BASE = 'https://skyview.gsfc.nasa.gov/cgi-bin/images?'
        params = f'Survey={survey}&position={position}&Size={fov}&Return=JPEG'
        self.url = BASE + params
        f = urlopen(self.url)
        
        self.data = plt.imread(f, format='jpeg')

    def plot(self):
        """
        Plots the image

        Returns
        -------
            fig, ax
        """
        fig, ax = plt.subplots()
        ax.imshow(self.data, cmap='gray')
        ax.set_xticks([])
        ax.set_yticks([])
        return fig, ax

    def show(self):
        """Shows the image"""
        fig, ax = self.plot()
        plt.show()

def plot_pm(ra, dec, pmra, pmdec, alpha=0.5, color=None):
    """
    Plot the proper motion of objects

    Arguments
    ---------
        ra (iterable of floats): Array of RA
        dec (iterable of floats): Array of DEC
        pmra (iterable of floats): Array of pmRA
        pmdec (iterable of floats): Array of pmDEC
        color (iterable of trings): Array of colors

    Returns
    -------
        fig, ax
    """
    origin = np.array([ra, dec])
    fig, ax = plt.subplots()
    ax.quiver(*origin, pmra, pmdec, alpha=alpha)
    ax.scatter(ra, dec, color=color)
    ax.set_aspect('equal')
    ax.set_xlabel('RA')
    ax.set_ylabel('DEC')
    return fig, ax
